from iobeam.utils import utils

'''
Represents a data query.
'''
class Query(object):

    def __init__(self, projectId, deviceId=None, seriesName=None):
        utils.checkValidProjectId(projectId)

        self._pid = projectId
        self._did = deviceId
        self._series = seriesName
        self._params = {}

    '''
    Gets the resource locator portion of the exports API call

    Returns:
        String of format <projectId>/<deviceId or 'all'>/<seriesName or 'all'>
    '''
    def getUrl(self):
        did = self._did if self._did is not None else "all"
        series = self._series if self._series is not None else "all"
        return "{}/{}/{}".format(self._pid, did, series)

    def getParams(self):
        return self._params

    '''
    Sets the limit of this query, i.e., the max number of results series.
    '''
    def limit(self, limit):
        if limit is None:
            return self
        if not isinstance(limit, int) or not limit > 0:
            raise ValueError("limit must be a positive int")

        self._params["limit"] = limit
        return self

    '''
    Sets the from time limit for results.
    '''
    def fromTime(self, time):
        if time is None:
            return self
        elif not isinstance(time, int):
            raise ValueError("time must be an int (milliseconds from epoch)")

        self._params["from"] = time
        return self

    '''
    Sets the to time limit for results.
    '''
    def toTime(self, time):
        if time is None:
            return self
        elif not isinstance(time, int):
            raise ValueError("time must be an int (milliseconds from epoch)")

        self._params["to"] = time
        return self

    '''
    Sets the max value for values in the results.
    '''
    def lessThan(self, value):
        if value is None:
            return self
        elif not isinstance(value, int):
            raise ValueError("value must be an int")

        self._params["less_than"] = value
        return self

    '''
    Sets the min value for values in the results.
    '''
    def greaterThan(self, value):
        if value is None:
            return self
        elif not isinstance(value, int):
            raise ValueError("value must be an int")

        self._params["greater_than"] = value
        return self

    '''
    Sets the value that all results must equal.
    '''
    def equals(self, value):
        if value is None:
            return self
        elif not isinstance(value, int):
            raise ValueError("value must be an int")

        self._params["equals"] = value
        return self



from iobeam.endpoints import service
from iobeam.http import request


'''
Communicates with the backend and exposes available Imports API methods.
'''
class ImportService(service.EndpointService):

    ''' Max number of points in a single request '''
    _BATCH_SIZE = 1000

    def __init__(self, token, requester=None):
        service.EndpointService.__init__(self, token, requester=requester)

    '''
    Creates the body of a import request.

    Params:
        projectId - Project ID of the request
        deviceId - Device ID of the request
        dataset - The data series of the request, a map with names to lists
            of data.DataPoints.

    Returns:
        A dictionary that is the body of an import request.
    '''
    @staticmethod
    def _makeRequest(projectId, deviceId, dataset):
        sources = []
        req = {
            "project_id": projectId,
            "device_id": deviceId,
            "sources": sources,
            "timefmt": "usec"
        }
        for series in dataset:
            pts = []
            obj = {"name": series, "data": pts}
            for d in dataset[series]:
                pts.append(d.toDict())
            sources.append(obj)

        return req

    '''
    Creates a list of import requests from a data set.

    If the data set is under _BATCH_SIZE, it will be one request. Otherwise
    it will be split into multiple requests as follows:
    (1) if a single series has less than ImportService._BATCH_SIZE points,
        it will be a request.
    (2) if a single series has more, it will be broken into multiple requests of
        ImportService._BATCH_SIZE size.

    Params:
        projectId - Project ID of the requests
        deviceId - Device ID of the requests
        dataset - The data set that will be broken into requests.

    Returns:
        A list of import request bodies.
    '''
    @staticmethod
    def _makeListOfReqs(projectId, deviceId, dataset):
        totalLen = sum(len(dataset[k]) for k in dataset)

        reqs = []
        # No series, no requests
        if totalLen == 0:
            pass
        # Everything can fit in one request
        elif totalLen <= ImportService._BATCH_SIZE:
            reqs.append(
                ImportService._makeRequest(projectId, deviceId, dataset))
        # Need to create multiple requests
        else:
            for series in dataset:
                seriesLen = len(dataset[series])
                # If the series itself is small enough, send it
                if seriesLen <= ImportService._BATCH_SIZE:
                    temp = {}
                    temp[series] = dataset[series]
                    reqs.append(
                        ImportService._makeRequest(projectId, deviceId, temp))
                # Split series into chunks of up to _BATCH_SIZE
                else:
                    idx = 0
                    valsList = list(dataset[series])
                    while idx < seriesLen:
                        end = idx + ImportService._BATCH_SIZE
                        vals = valsList[idx:end]
                        temp = {}
                        temp[series] = vals
                        reqs.append(ImportService._makeRequest(
                            projectId, deviceId, temp))
                        idx += ImportService._BATCH_SIZE

        return reqs

    '''
    Wraps API call `POST /imports`

    Sends data to the iobeam backend to be stored.

    Params:
        projectId - Project ID the data belongs to
        deviceId - Device ID the data belongs to
        dataSeries - Dataset to send, as a dictionary where the keys are
            the name of the series, and the values are sets containing
            `iobeam.iobeam.DataPoint`s.

    Returns:
        True if the data is sent successfully; False and the response otherwise.

    Raises:
        Exception - If any of projectId, deviceId, or dataSeries is None.
    '''
    def importData(self, projectId, deviceId, dataSeries):
        if not self.token:
            raise request.UnauthorizedError.noTokenSet();
        if projectId is None:
            raise Exception("Project ID cannot be None")
        elif deviceId is None:
            raise Exception("Device ID cannot be None")
        elif dataSeries is None:
            raise Exception("Dataset cannot be None")
        elif len(dataSeries) == 0:
            return True
        endpoint = self.makeEndpoint("imports")

        reqs = ImportService._makeListOfReqs(projectId, deviceId, dataSeries)
        success = True
        extra = None
        for req in reqs:
            r = self.requester().post(endpoint).token(self.token)
            r.setBody(req)
            r.execute()
            success = success and (r.getResponseCode() == 200)
            if r.getResponseCode() != 200:
                extra = r.getResponse()

        return (success, extra)



from time import time
from enum import Enum

class TimeUnit(Enum):
    MILLISECONDS = 0
    MICROSECONDS = 1
    SECONDS = 2

class Timestamp(object):

    def __init__(self, value, type=TimeUnit.MILLISECONDS):
        self._value = value
        self._type = type

    def __eq__(self, other):
        if other is None or not isinstance(other, Timestamp):
            return False

        return self.asMicroseconds() == other.asMicroseconds()

    def asMicroseconds(self):
        if self._type == TimeUnit.MICROSECONDS:
            return self._value
        elif self._type == TimeUnit.MILLISECONDS:
            return int(self._value * 1000)
        elif self._type == TimeUnit.SECONDS:
            return int(self._value * 1000000)
        else:
            raise ValueError("unknown type")

'''
Represents a time-series datapoint, using a timestamp and value.
'''
class DataPoint(object):

    '''
    Constructor of a DataPoint.

    Takes in a value, which must be a number type, and optionally a
    timestamp. If a timestamp is not supplied, the current time in
    milliseconds is used.

    Raises:
        ValueError - If value is not a number (int/float) or timestamp is not an
                     int.
    '''
    def __init__(self, value, timestamp=None):
        if not (isinstance(value, int) or isinstance(value, float)):
            raise ValueError("'value' must be a number.")
        if timestamp is None:
            self._timestamp = Timestamp(int(time() * 1000),
                                        TimeUnit.MILLISECONDS)
        elif isinstance(timestamp, int):
            self._timestamp = Timestamp(timestamp, TimeUnit.MILLISECONDS)
        elif isinstance(timestamp, Timestamp):
            self._timestamp = timestamp
        else:
            raise ValueError(
                "invalid type for timestamp: {}".format(type(timestamp)))

        self._value = value

    '''
    Prints out: DataPoint{timestamp: <timestamp>, value: <value>}
    '''
    def __str__(self):
        return "DataPoint{{timestamp: {}, value: {}}}".format(
            self._timestamp.asMicroseconds(), self._value)

    def __eq__(self, other):
        if other is None or not isinstance(other, DataPoint):
            return False
        return (self._value == other._value) and (
            self._timestamp == other._timestamp)

    def __hash__(self):
        return self._timestamp.asMicroseconds() + self._value

    '''
    Converts to a dictionary that is usable for the imports API.

    Returns:
        Dictionary with keys 'time' and 'value'.
    '''
    def toDict(self):
        return {
            "time": self._timestamp.asMicroseconds(),
            "value": self._value
        }

'''
A collection of DataPoints for a given named series.
'''
class DataSeries(object):

    '''
    Constructor for DataSeries with a name and its points.

    Params:
        name - Name of the data series
        points - DataPoints in the series

    Throws:
        ValueError - If `name` is None
    '''
    def __init__(self, name, points):
        if name is None:
            raise ValueError("Name cannot be None")
        self._name = name
        self._points = list(points) if points is not None else []

    '''
    Returns the number of points in the series
    '''
    def __len__(self):
        return len(self._points)

    '''
    Prints out: DataSeries{name: <name>, len: <size of points>}
    '''
    def __str__(self):
        return "DataSeries{{name: {}, len: {}}}".format(self._name, len(self))

    def getName(self):
        return self._name

    def getPoints(self):
        return self._points

'''
Creates a DataSeries from a list of values at uniform time intervals.

Given a `startTime` and `endTime`, this function creates a DataSeries where
each DataPoint is separated by a uniform time interval, with the first point
at `startTime` and the last one at `endTime`. For example, if
`startTime` is 0 and `endTime is 100, and there are 5 values, the resulting
DataPoints would be at times 0, 25, 50, 75, 100.

Params:
    seriesName - Name of the resulting DataSeries
    startTime - Time in milliseconds when the series should start
    endTime - Time in milliseconds when the series should end
    values - Numerical values to be made into DataPoints.

Returns:
    A uniformly distributed DataSeries with DataPoints corresponding to
    `values`. None if the parameters are invalid, i.e., None or
    endTime <= startTime.
'''
def makeUniformDataSeries(seriesName, startTime, endTime, values):
    if (seriesName is None) or (startTime is None) or (endTime is None):
        return None
    elif not isinstance(startTime, int) or not isinstance(endTime, int):
        return None  # TODO: should raise ValueError?
    elif endTime <= startTime:
        return None  # TODO: should raise ValueError?
    elif (values is None) or len(values) == 0:
        return DataSeries(seriesName, [])

    valLen = len(values)
    if valLen == 1:
        pt = DataPoint(values[0], startTime)
        return DataSeries(seriesName, [pt])
    elif valLen == 2:
        pt1 = DataPoint(values[0], startTime)
        pt2 = DataPoint(values[1], endTime)
        return DataSeries(seriesName, [pt1, pt2])

    interval = (endTime - startTime) / (valLen - 1)
    pts = []
    # For compatibility with both Python 2 and 3.
    try:
        xrange
    except NameError:
        xrange = range  # Python3
    for i in xrange(0, valLen - 1):
        t = int(startTime + (i * interval))
        d = DataPoint(values[i], timestamp=t)
        pts.append(d)
    pts.append(DataPoint(values[valLen - 1], timestamp=endTime))

    return DataSeries(seriesName, pts)



from iobeam.utils import utils

'''
Represents a device that is associated with a project.
'''
class Device(object):

    '''
    Constructor for a Device object.

    A valid Device object has at least a `projectId` and a `deviceId`, and
    optionally a `deviceName`.

    Params:
        projectId - Project id (int) that this device belongs to
        deviceId - Id of this device
        deviceName - Optional secondary identifier for the device

    Raises:
        ValueError - If projectId is None, not an int, or not >0. Also if
                     deviceId is None.
    '''
    def __init__(self, projectId, deviceId, deviceName=None):
        utils.checkValidProjectId(projectId)
        utils.checkValidDeviceId(deviceId)

        self.projectId = projectId
        self.deviceId = deviceId
        self.deviceName = deviceName

    '''
    Prints a string representation of this device in the form:
    <deviceName> [<projectId>: <deviceId>]
    '''
    def __str__(self):
        return "{} [{}: {}]".format(self.deviceName, self.projectId,
                                    self.deviceId)



from .endpoints import devices
from .endpoints import exports
from .endpoints import imports
from .http import request
from .resources import data
from .resources import device
from .resources import query
from .utils import utils
import os.path

'''
Aliases for resource types for convenience outside the package.
'''
DataPoint = data.DataPoint
DataSeries = data.DataSeries
Timestamp = data.Timestamp
TimeUnit = data.TimeUnit
QueryReq = query.Query

_DEVICE_ID_FILE = "iobeam_device_id"

class ClientBuilder(object):

    def __init__(self, projectId, projectToken):
        utils.checkValidProjectId(projectId)
        utils.checkValidProjectToken(projectToken)

        self._projectId = projectId
        self._projectToken = projectToken
        self._diskPath = None
        self._deviceId = None
        self._regArgs = None
        self._backend = None

    def saveToDisk(self, path="."):
        self._diskPath = path
        return self

    def setDeviceId(self, deviceId):
        utils.checkValidDeviceId(deviceId)
        self._deviceId = deviceId
        return self

    def registerDevice(self, deviceId=None, deviceName=None):
        self._regArgs = (deviceId, deviceName)
        return self

    def setBackend(self, baseUrl):
        self._backend = request.Requester(baseUrl=baseUrl)
        return self

    def build(self):
        client = _Client(self._diskPath, self._projectId, self._projectToken,
                         self._backend, deviceId=self._deviceId)
        if (self._regArgs is not None):
            did, dname = self._regArgs
            client.registerDevice(deviceId=did, deviceName=dname)

        return client


class _Client(object):

    '''
    Constructor for iobeam client object.

    Creates a client instance associated with a project and (potentially) a
    device. If `path` is provided, this device's ID will be stored at
    <path>/iobeam_device_id. This on-disk ID will be used if one is not
    provided as `deviceId`.

    Params:
        path - Path where device ID should be persisted
        projectId - iobeam project ID
        projectToken - iobeam project token with write access for sending data
        backend - Base url of the backend to use; if None, requests go to
                  https://api.iobeam.com/v1/
        deviceId - Device id if previously registered
    '''
    def __init__(self, path, projectId, projectToken, backend, deviceId=None):
        utils.checkValidProjectId(projectId)
        utils.checkValidProjectToken(projectToken)

        self.projectId = projectId
        self.projectToken = projectToken
        self._path = path
        self._dataset = {}

        self._activeDevice = None
        if deviceId is not None:
            self._setActiveDevice(device.Device(projectId, deviceId))
        elif self._path is not None:
            p = os.path.join(self._path, _DEVICE_ID_FILE)
            if os.path.isfile(p):
                with open(p, "r") as f:
                    did = f.read()
                    if len(did) > 0:
                        self._activeDevice = device.Device(projectId, did)


        # Setup services
        requester = backend
        self._deviceService = devices.DeviceService(projectToken,
                                                    requester=backend)
        self._importService = imports.ImportService(projectToken,
                                                    requester=backend)

    '''
    Registers the device with iobeam.

    If a path was provided when the client was constructed, the device ID
    will be stored on disk.

    Params:
        deviceId - Desired device ID; otherwise randomly generated
        deviceName - Desired device name; otherwise randomly generated

    Returns:
        This client object (allows for chaining)
    '''
    def registerDevice(self, deviceId=None, deviceName=None):
        if self._activeDevice is not None and (deviceId is None or
            self._activeDevice.deviceId == deviceId):
            return self

        d = self._deviceService.registerDevice(self.projectId,
            deviceId=deviceId, deviceName=deviceName)
        self._setActiveDevice(d)

        return self

    def setDeviceId(self, deviceId):
        d = device.Device(self.projectId, deviceId)
        self._setActiveDevice(d)

    def getDeviceId(self):
        if self._activeDevice is None:
            return None
        else:
            return self._activeDevice.deviceId

    def _setActiveDevice(self, device):
        self._activeDevice = device
        if self._path is not None:
            p = os.path.join(self._path, _DEVICE_ID_FILE)
            with open(p, "w") as f:
                f.write(self._activeDevice.deviceId)

    '''
    Tells whether this client has a registered device.

    Returns:
        True if there is a device ID associated with this client
    '''
    def isRegistered(self):
        return self._activeDevice is not None

    '''
    Adds a single DataPoint to a series in the data store.

    If the series does not currently exist, it will be created.

    Params:
        seriesName - The series to add the datapoint to
        datapoint - DataPoint to be added
    '''
    def addDataPoint(self, seriesName, datapoint):
        if seriesName is None or not isinstance(seriesName, str):
            return
        elif len(seriesName) == 0:
            return
        elif datapoint is None or not isinstance(datapoint, data.DataPoint):
            return

        if seriesName not in self._dataset:
            self._dataset[seriesName] = set()
        self._dataset[seriesName].add(datapoint)

    '''
    Adds a DataSeries to the data store.

    If the series exists, all the points will be added.

    Params:
        dataseries - The DataSeries to add to the data store.
    '''
    def addDataSeries(self, dataseries):
        if dataseries is None or not isinstance(dataseries, data.DataSeries):
            return
        elif len(dataseries) == 0:
            return

        key = dataseries.getName()
        if key not in self._dataset:
            self._dataset[key] = set()

        for p in dataseries.getPoints():
            self._dataset[key].add(p)

    '''
    Removes any points associated with `seriesName`.

    Params:
        seriesName - Name of the data series to be cleared.
    '''
    def clearSeries(self, seriesName):
        self._dataset.pop(seriesName, None)

    '''
    Sends stored data to the iobeam backend.

    Raises:
        Will throw an Exception if sending the data fails.
    '''
    def send(self):
        pid = self.projectId
        did = self._activeDevice.deviceId
        dataset = self._dataset
        success, extra = self._importService.importData(pid, did, dataset)
        if not success:
            raise Exception("Send failed, server sent: {}".format(extra))

    '''
    Performs a query on the iobeam backend.

    The Query specifies the project, device, and series to look up, as well
    as any parameters to use.

    Params:
        token - A token with read access for the given project.
        query - Specifies a data query to perform.

    Returns:
        A dictionary representing the results of the query.

    Raises:
        ValueError - If `token` or `query` is None, or `query` is the wrong
                     type.
    '''
    @staticmethod
    def query(token, query, backend=None):
        if token is None:
            raise ValueError("token cannot be None")
        elif query is None:
            raise ValueError("query cannot be None")
        elif not isinstance(query, QueryReq):
            raise ValueError("query must be a iobeam.QueryReq")
        requester = None
        if backend is not None:
            requester = request.Requester(baseUrl=backend)

        print(requester)
        service = exports.ExportService(token, requester=requester)
        return service.getData(query)

MakeQuery = _Client.query
